use core::fmt::{Debug, Display};

use alloc::string::String;
use alloc::vec::Vec;
use serde::{Deserialize, Serialize};

use super::{
    errno::Errno,
    message::DissectedMessage,
    rpc_header::{RpcHeader, RpcMessageType},
};

#[derive(Serialize, Deserialize, Debug)]
pub struct WaitResult {
    pub pid: u64,
    pub exit_code: u8,
}

#[derive(Clone, Copy, Debug, Serialize, Deserialize)]
pub enum ProcessStatus {
    Running,
    Exited(u8),
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Process {
    pub pid: u64,
    pub stdio: [Option<(u64, u64)>; 3],
    pub cli_args: Vec<String>,
    pub status: ProcessStatus,
    pub parent: Option<u64>,
    pub children: Vec<u64>,
}

#[derive(Deserialize)]
pub struct OsString {
    #[expect(unused, reason = "this is a dummy for deserialization")]
    kind: u8,
    str: String,
}

impl Debug for OsString {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.write_fmt(format_args!("{:?}", self.str))
    }
}

pub enum ProcManRpcCall {
    NewProc { pid: u64, parent: Option<u64> },
    SetStdio { pid: u64, stdio: [Option<(u64, u64)>; 3] },
    GetStdio,
    SetCliArgs { pid: u64, args: Vec<OsString> },
    GetCliArgs,
    Exit { pid: u64, code: u8 },
    Wait { pid: u64, block: bool },
    GetProcesses,
    SetCwd { pid: u64, path: String },
    GetCwd,
    Invalid,
}

pub enum ProcManRpcReturn {
    NewProc(Result<(), Errno>),
    SetStdio(Result<(), Errno>),
    GetStdio([Option<(u64, u64)>; 3]),
    SetCliArgs(Result<(), Errno>),
    GetCliArgs(Vec<OsString>),
    Wait(Result<WaitResult, Errno>),
    GetProcesses(
        #[expect(unused, reason = "currently we do not use the full parsed message in output")]
        Vec<Process>,
    ),
    SetCwd(Result<(), Errno>),
    GetCwd(String),
    Invalid,
}

pub enum ProcManRpcMessage {
    Call(ProcManRpcCall),
    Return(ProcManRpcReturn),
}

impl ProcManRpcMessage {
    pub fn dissect(buffer: &[u8], message: &DissectedMessage<RpcHeader>) -> Self {
        let rpc_header = message.get_l1();
        match rpc_header.typ {
            RpcMessageType::Call => {
                let call = match rpc_header.func {
                    0 => {
                        let (pid, stdio) = postcard::from_bytes(buffer).unwrap();
                        ProcManRpcCall::SetStdio { pid, stdio }
                    }
                    1 => ProcManRpcCall::GetStdio,
                    2 => {
                        let (pid, args) = postcard::from_bytes(buffer).unwrap();
                        ProcManRpcCall::SetCliArgs { pid, args }
                    }
                    3 => ProcManRpcCall::GetCliArgs,
                    4 => {
                        let (pid, code) = postcard::from_bytes(buffer).unwrap();
                        ProcManRpcCall::Exit { pid, code }
                    }
                    5 => {
                        let (pid, block) = postcard::from_bytes(buffer).unwrap();
                        ProcManRpcCall::Wait { pid, block }
                    }
                    6 | 7 => {
                        let (pid, parent) = postcard::from_bytes(buffer).unwrap();
                        ProcManRpcCall::NewProc { pid, parent }
                    }
                    8 => ProcManRpcCall::GetProcesses,
                    9 => {
                        let (pid, path) = postcard::from_bytes(buffer).unwrap();
                        ProcManRpcCall::SetCwd { pid, path }
                    }
                    10 => ProcManRpcCall::GetCwd,
                    _ => ProcManRpcCall::Invalid,
                };
                Self::Call(call)
            }
            RpcMessageType::Return => {
                let call = match rpc_header.func {
                    0 => ProcManRpcReturn::SetStdio(postcard::from_bytes(buffer).unwrap()),
                    1 => ProcManRpcReturn::GetStdio(postcard::from_bytes(buffer).unwrap()),
                    2 => ProcManRpcReturn::SetCliArgs(postcard::from_bytes(buffer).unwrap()),
                    3 => ProcManRpcReturn::GetCliArgs(postcard::from_bytes(buffer).unwrap()),
                    5 => ProcManRpcReturn::Wait(postcard::from_bytes(buffer).unwrap()),
                    7 => ProcManRpcReturn::NewProc(postcard::from_bytes(buffer).unwrap()),
                    8 => ProcManRpcReturn::GetProcesses(postcard::from_bytes(buffer).unwrap()),
                    9 => ProcManRpcReturn::SetCwd(postcard::from_bytes(buffer).unwrap()),
                    10 => ProcManRpcReturn::GetCwd(postcard::from_bytes(buffer).unwrap()),
                    _ => ProcManRpcReturn::Invalid,
                };
                Self::Return(call)
            }
        }
    }
}

impl Display for ProcManRpcMessage {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            Self::Call(call) => match call {
                ProcManRpcCall::NewProc { pid, parent } => {
                    if let Some(parent) = parent {
                        f.write_fmt(format_args!(
                            "New process with PID {pid} and parent {parent}"
                        ))?;
                    } else {
                        f.write_fmt(format_args!("New process with PID {pid} and no parent"))?;
                    }
                }
                ProcManRpcCall::SetStdio { pid, stdio } => {
                    f.write_fmt(format_args!("Set PID {pid}'s stdio to {stdio:?}"))?;
                }
                ProcManRpcCall::GetStdio => f.write_str("Get stdio")?,
                ProcManRpcCall::SetCliArgs { pid, args } => {
                    f.write_fmt(format_args!("Set PID {pid}'s CLI args to {args:?}"))?;
                }
                ProcManRpcCall::GetCliArgs => f.write_str("Get CLI args")?,
                ProcManRpcCall::Exit { pid, code } => {
                    f.write_fmt(format_args!("PID {pid} exiting with code {code}"))?;
                }
                ProcManRpcCall::Wait { pid, block } => {
                    if *block {
                        f.write_fmt(format_args!("Wait blocking on PID {pid}"))?;
                    } else {
                        f.write_fmt(format_args!("Wait nonblocking on PID {pid}"))?;
                    }
                }
                ProcManRpcCall::GetProcesses => f.write_str("Get processes")?,
                ProcManRpcCall::SetCwd { pid, path } => {
                    f.write_fmt(format_args!("Set PID {pid}'s CWD to {path}"))?;
                }
                ProcManRpcCall::GetCwd => f.write_str("Get CWD")?,
                ProcManRpcCall::Invalid => f.write_str("Invalid call!")?,
            },
            Self::Return(ret) => match ret {
                ProcManRpcReturn::NewProc(ret) => {
                    f.write_fmt(format_args!("New process return {ret:?}"))?;
                }
                ProcManRpcReturn::SetStdio(ret) => {
                    f.write_fmt(format_args!("Set stdio return {ret:?}"))?;
                }
                ProcManRpcReturn::GetStdio(ret) => {
                    f.write_fmt(format_args!("Get stdio return {ret:?}"))?;
                }
                ProcManRpcReturn::SetCliArgs(ret) => {
                    f.write_fmt(format_args!("Set CLI args return {ret:?}"))?;
                }
                ProcManRpcReturn::GetCliArgs(ret) => {
                    f.write_fmt(format_args!("Get CLI args return {ret:?}"))?;
                }
                ProcManRpcReturn::Wait(ret) => f.write_fmt(format_args!("Wait return {ret:?}"))?,
                ProcManRpcReturn::GetProcesses(_) => f.write_str("Get processes return")?,
                ProcManRpcReturn::SetCwd(ret) => {
                    f.write_fmt(format_args!("Set CWD return {ret:?}"))?;
                }
                ProcManRpcReturn::GetCwd(ret) => {
                    f.write_fmt(format_args!("Get CWD return {ret:?}"))?;
                }
                ProcManRpcReturn::Invalid => f.write_str("Invalid return!")?,
            },
        };
        Ok(())
    }
}
