use core::fmt::Display;

use alloc::string::String;
use alloc::vec::Vec;

use super::{
    message::DissectedMessage,
    rpc_header::{RpcHeader, RpcMessageType},
    syslog_msg::SyslogMessage,
};

pub enum SyslogRpcCall {
    SendMessage { message: SyslogMessage },
    SubscribeToText,
    SubscribeToBinary { message_from: String, kinds: Vec<u64> },
    Invalid,
}

pub enum SyslogRpcReturn {
    SendMessage(Result<(), ()>),
    SubscribeToText,
    SubscribeToBinary,
    Invalid,
}

pub enum SyslogRpcMessage {
    Call(SyslogRpcCall),
    Return(SyslogRpcReturn),
}

impl SyslogRpcMessage {
    pub fn dissect(buffer: &[u8], message: &DissectedMessage<RpcHeader>) -> Self {
        let rpc_header = message.get_l1();
        match rpc_header.typ {
            RpcMessageType::Call => {
                let call = match rpc_header.func {
                    0 => {
                        let message = postcard::from_bytes(buffer).unwrap();
                        SyslogRpcCall::SendMessage { message }
                    }
                    1 => SyslogRpcCall::SubscribeToText,
                    2 => {
                        let (message_from, kinds) = postcard::from_bytes(buffer).unwrap();
                        SyslogRpcCall::SubscribeToBinary { message_from, kinds }
                    }
                    _ => SyslogRpcCall::Invalid,
                };
                Self::Call(call)
            }
            RpcMessageType::Return => {
                let call = match rpc_header.func {
                    0 => SyslogRpcReturn::SendMessage(postcard::from_bytes(buffer).unwrap()),
                    1 => SyslogRpcReturn::SubscribeToText,
                    2 => SyslogRpcReturn::SubscribeToBinary,
                    _ => SyslogRpcReturn::Invalid,
                };
                Self::Return(call)
            }
        }
    }
}

impl Display for SyslogRpcMessage {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            Self::Call(call) => match call {
                SyslogRpcCall::SendMessage { message } => {
                    f.write_fmt(format_args!("{message}"))?;
                }
                SyslogRpcCall::SubscribeToText => f.write_str("Subscribe to text")?,
                SyslogRpcCall::SubscribeToBinary { message_from, kinds } => {
                    f.write_fmt(format_args!(
                        "Subscribe to binary from {message_from} with kinds {kinds:?}",
                    ))?;
                }
                SyslogRpcCall::Invalid => f.write_str("Invalid call!")?,
            },
            Self::Return(ret) => match ret {
                SyslogRpcReturn::SendMessage(ret) => {
                    f.write_fmt(format_args!("Send message return {ret:?}"))?;
                }
                SyslogRpcReturn::SubscribeToText => f.write_str("Subscribe to text return")?,
                SyslogRpcReturn::SubscribeToBinary => f.write_str("Subscribe to binary return")?,
                SyslogRpcReturn::Invalid => f.write_str("Invalid return!")?,
            },
        };
        Ok(())
    }
}
