use core::sync::atomic::{AtomicBool, AtomicU64, Ordering};

use crate::{
    interrupts::{self, EoiGuard},
    println,
    tasking::Tid,
    TASKING,
};
use alloc::vec::Vec;
use spin::Mutex;
use x86_64::instructions::port::{Port, PortWriteOnly};

static DATA: Mutex<Port<u8>> = Mutex::new(Port::new(0x40));
static CMD: Mutex<PortWriteOnly<u8>> = Mutex::new(PortWriteOnly::new(0x43));

const MAX_FREQ: u32 = 1_193_180;

pub static NUM_INTERRUPTS: AtomicU64 = AtomicU64::new(0);

pub static THREAD_SLEEP_LIST: Mutex<Vec<(u64, Tid)>> = Mutex::new(Vec::new());
pub static CHECK_SLEEP_LIST: AtomicBool = AtomicBool::new(false);

pub fn init(mut freq: u32) {
    assert_ne!(freq, 0);
    #[expect(
        clippy::unwrap_used,
        reason = "register_handler requires the interrupt number to be less than 16, which it is."
    )]
    interrupts::register_handler(0, handler).unwrap();
    #[expect(clippy::arithmetic_side_effects, reason = "freq has been checked to not be 0")]
    let mut div = MAX_FREQ / freq;
    if div > 65535 {
        println!("[PIT] Frequency of {}Hz too slow, min freq is 18 Hz", freq);
        div = 65535;
        freq = 18;
    } else if div == 0 {
        println!("[PIT] Frequency of {}Hz too fast, max freq is {} Hz", freq, MAX_FREQ);
        div = 1;
        freq = MAX_FREQ;
    }
    #[expect(
        clippy::unwrap_used,
        reason = "div has been checked to be in u16 range above, so this cannot panic"
    )]
    let div_bytes = u16::try_from(div).unwrap().to_le_bytes();

    println!("[PIT] Setting PIT to {}Hz with divisor of {}", freq, div);
    // Command breakdown (MSB to LSB):
    // 00 - Channel 0
    // 11 - lobyte/hibyte access mode - set both bytes in one command
    // 011 - Mode 3, sqaure wave generator at frequency MAX_FREQ/div, generates interrupts at that
    //     frequency
    // 0 - binary mode, always used
    unsafe { CMD.lock().write(0b0011_0110_u8) };
    unsafe { DATA.lock().write(div_bytes[0]) };
    unsafe { DATA.lock().write(div_bytes[1]) };
    CHECK_SLEEP_LIST.store(true, Ordering::Relaxed);
}

fn handler(_irq: u8, eoi_guard: EoiGuard) {
    let ticks = NUM_INTERRUPTS.fetch_add(1, Ordering::Relaxed) + 1;
    if CHECK_SLEEP_LIST.load(Ordering::Relaxed) {
        let mut list = THREAD_SLEEP_LIST.lock();
        while list.last().is_some_and(|x| x.0 <= ticks) {
            let entry = list.pop().unwrap();
            TASKING.force_wake(entry.1).unwrap();
        }
    }
    drop(eoi_guard);
    if TASKING.ok_to_yield() {
        TASKING.task_yield();
    }
}

pub fn thread_sleeping(tid: Tid, duration: u64) {
    let end = NUM_INTERRUPTS.load(Ordering::Relaxed) + duration;
    let mut list = THREAD_SLEEP_LIST.lock();
    list.push((end, tid));
    list.sort_by(|x, y| (y.0).cmp(&x.0));
}

pub fn cancel_sleep(tid: Tid) {
    THREAD_SLEEP_LIST.lock().retain(|x| x.1 != tid);
}
