use core::{fmt, fmt::Write};
use spin::{Lazy, Mutex};
use uart_16550::SerialPort;

pub static FIRST_PORT: Lazy<Wrapper> = Lazy::new(|| {
    // SAFETY:
    // 0x3f8 is the defined address for the first serial port on x86_64,
    // so we know it is safe to use that as the port's base address.
    let mut port = unsafe { SerialPort::new(0x3f8) };
    port.init();
    Wrapper(Mutex::new(port))
});

#[cfg(any(feature = "log-rpc-pcapng", feature = "log-rpc-text"))]
pub static SECOND_PORT: Lazy<Wrapper> = Lazy::new(|| {
    // SAFETY:
    // 0x2f8 is the defined address for the second serial port on x86_64,
    // so we know it is safe to use that as the port's base address.
    let mut port = unsafe { SerialPort::new(0x2f8) };
    port.init();
    Wrapper(Mutex::new(port))
});

pub struct Wrapper(Mutex<SerialPort>);
impl fmt::Write for &Wrapper {
    fn write_str(&mut self, s: &str) -> fmt::Result {
        self.0.lock().write_str(s)
    }
}

impl Wrapper {
    pub fn write_bytes(&self, bytes: &[u8]) {
        let mut port = self.0.lock();
        for byte in bytes {
            port.send_raw(*byte);
        }
    }

    pub fn write_u32s(&self, u32s: &[u32]) {
        for elem in u32s {
            let bytes = elem.to_ne_bytes();
            self.write_bytes(&bytes);
        }
    }
}

#[macro_export]
macro_rules! print {
    ($($arg:tt)*) => ({
        #[expect(clippy::used_underscore_items, reason="_print is meant for this macro")]
        $crate::serial::_print(format_args!($($arg)*))
    })
}

#[macro_export]
macro_rules! println {
    () => ($crate::print!("\r\n"));
    ($($arg:tt)*) => ({
        #[expect(clippy::used_underscore_items, reason="_print is meant for this macro")]
        $crate::serial::_print(format_args!($($arg)*));
        $crate::print!("\r\n");
    })
}

#[macro_export]
macro_rules! dbg {
    () => {
        $crate::println!("[{}:{}]", file!(), line!())
    };
    ($val:expr $(,)?) => {
        match $val {
            tmp => {
                $crate::println!("[{}:{}] {} = {:#?}",
                    file!(), line!(), stringify!($val), &tmp);
                tmp
            }
        }
    };
    ($($val:expr),+ $(,)?) => {
        ($($crate::dbg!($val)),+,)
    };
}

#[doc(hidden)]
pub fn _print(args: fmt::Arguments) {
    #[expect(
        clippy::expect_used,
        reason = "Kernel prints tend to be fairly important, so if printing fails a panic is in order."
    )]
    (&*FIRST_PORT).write_fmt(args).expect("Failed to print to the serial port");
}
